@@ -100,18 +100,23 @@ testNumerics <- function(numbers, positives=NULL, integers=NULL,
 
 #' Remove related individuals while keeping maximum number of individuals
 #'
-#' \code{filterRelatedness} takes a data.frame with pair-wise relatedness
+#' \code{relatednessFilter} takes a data.frame with pair-wise relatedness
 #' measures of samples and returns pairs of individual IDs that are related as
 #' well as a list of suggested individual IDs to remove.
-#' \code{filterRelatedness} finds pairs of samples whose relatedness estimate is
-#' larger than the specified relatednessTh. Subsequently, for pairs of
+#' \code{relatednessFilter} finds pairs of samples whose relatedness estimate
+#' is larger than the specified relatednessTh. Subsequently, for pairs of
 #' individual that do not have additional relatives in the dataset, the
 #' individual with the worse otherCriterionMeasure (if provided) or arbitrarily
 #' individual 1 of that pair is selected and returned as the individual failing
 #' the relatedness check. For more complex family structures, the unrelated
-#' individuals per family are selected (e.g. in a parents-offspring trio, the
-#' offspring will be marked as fail, while the parents will be kept in the
-#' analysis).
+#' individuals per family are selected (e.g. in a simple case of a
+#' parents-offspring trio, the offspring will be marked as fail, while the
+#' parents will be kept in the analysis). Selection is achieved by constructing
+#' subgraphs of clusters of individuals that are related.
+#' \code{relatednessFilter} then finds the maximum independent set of vertices
+#' in the subgraphs of related individuals. If all individuals are related (i.e.
+#' all maximum independent sets are 0), one individual of that cluster will be
+#' kept and all others listed as failIDs.
 #' @param relatedness [data.frame] containing pair-wise relatedness estimates
 #' (in column [relatednessRelatedness]) for individual 1 (in column
 #' [relatednessIID1] and individual 2 (in column [relatednessIID1]). Columns
@@ -155,11 +160,14 @@ testNumerics <- function(numbers, positives=NULL, integers=NULL,
 #' individual IDs.
 #' @param otherCriterionMeasure [character] Column name of the column containing
 #' the measure of the otherCriterion (for instance SNP missingness rate).
+#' @param verbose [logical] If TRUE, progress info is printed to standard out.
 #' @return named [list] with i) relatednessFails, a [data.frame] containing the
-#' reordered input data.frame relatedness after filtering for pairs of
-#' individuals in relatednessIID1 and relatednessIID2, that fail the relatedness
-#' QC and ii) failIDs, a [data.frame] with the [IID]s (and [FID]s if provided)
-#' of the individuals that fail the relatednessTh.
+#' data.frame relatedness after filtering for pairs of individuals in
+#' relatednessIID1 and relatednessIID2, that fail the relatedness
+#' QC; the data.frame is reordered with the fail individuals in column 1 and
+#' their related individuals in column 2 and ii) failIDs, a [data.frame] with
+#' the [IID]s (and [FID]s if provided) of the individuals that fail the
+#' relatednessTh.
 #' @export
 
 relatednessFilter <- function(relatedness, otherCriterion=NULL,
@@ -170,22 +178,33 @@ relatednessFilter <- function(relatedness, otherCriterion=NULL,
                               relatednessFID1=NULL, relatednessFID2=NULL,
                               relatednessRelatedness="PI_HAT",
                               otherCriterionIID="IID",
-                              otherCriterionMeasure=NULL ) {
+                              otherCriterionMeasure=NULL,
+                              verbose=FALSE) {
 
     if (!(relatednessIID1 %in% names(relatedness))) {
-        stop(paste("Column", relatednessIID1, "not found!"))
+        stop(paste("Column", relatednessIID1, "for relatedness not found!"))
     }
     if (!(relatednessIID2 %in% names(relatedness))) {
-        stop(paste("Column", relatednessIID1, "not found!"))
+        stop(paste("Column", relatednessIID1, "for relatedness not found!"))
     }
     if (!(relatednessRelatedness %in% names(relatedness))) {
-        stop(paste("Column", relatednessRelatedness, "not found!"))
+        stop(paste("Column", relatednessRelatedness,
+                   "for relatedness not found!"))
     }
+
+    iid1_index <- which(colnames(relatedness) == relatednessIID1)
+    iid2_index <- which(colnames(relatedness) == relatednessIID2)
+
+    relatedness[,iid1_index] <- as.character(relatedness[,iid1_index])
+    relatedness[,iid2_index] <- as.character(relatedness[,iid2_index])
+
     relatedness_names <- names(relatedness)
-    names(relatedness)[names(relatedness) == relatednessIID1] <- "IID1"
-    names(relatedness)[names(relatedness) == relatednessIID2] <- "IID2"
+    names(relatedness)[iid1_index] <- "IID1"
+    names(relatedness)[iid2_index] <- "IID2"
     names(relatedness)[names(relatedness) == relatednessRelatedness] <- "M"
 
+    relatedness_original <- relatedness
+
     if (!is.null(relatednessFID1) && is.null(relatednessFID2) ||
         is.null(relatednessFID1) && !is.null(relatednessFID2) ) {
         stop("Either none or both, relatednessFID1 and relatednessFID2 have to
@@ -193,36 +212,58 @@ relatednessFilter <- function(relatedness, otherCriterion=NULL,
     }
     if (!is.null(relatednessFID1) && !is.null(relatednessFID2)) {
         if (!(relatednessFID1 %in% names(relatedness))) {
-            stop(paste("Column", relatednessFID1, "not found!"))
+            stop(paste("Column", relatednessFID1, "for relatedness not found!"))
         }
         if (!(relatednessFID2 %in% names(relatedness))) {
-            stop(paste("Column", relatednessFID2, "not found!"))
+            stop(paste("Column", relatednessFID2, "for relatedness not found!"))
         }
-        names(relatedness)[names(relatedness) == relatednessFID1] <- "FID1"
-        names(relatedness)[names(relatedness) == relatednessFID2] <- "FID2"
+        fid1_index <- which(colnames(relatedness) == relatednessFID1)
+        fid2_index <- which(colnames(relatedness) == relatednessFID2)
+        names(relatedness)[fid1_index] <- "FID1"
+        names(relatedness)[fid2_index] <- "FID2"
+        relatedness$FID1 <- as.character(relatedness$FID1)
+        relatedness$FID2 <- as.character(relatedness$FID2)
     }
-    relatedness_original <- relatedness
+
+    # Remove symmetric IID rows
     relatedness <- dplyr::select_(relatedness, ~IID1, ~IID2, ~M)
 
+    sortedIDs <- data.frame(t(apply(relatedness, 1, function(pair) {
+        c(sort(c(pair[1], pair[2])))
+        })), stringsAsFactors=FALSE)
+    keepIndex <- which(!duplicated(sortedIDs))
+
+    relatedness_original <- relatedness_original[keepIndex,]
+    relatedness <- relatedness[keepIndex,]
 
     # individuals with at least one pair-wise comparison > relatednessTh
     highRelated <- dplyr::filter_(relatedness, ~M > relatednessTh)
+    if (nrow(highRelated) == 0) {
+        return(list(relatednessFails=NULL, failIDs=NULL))
+    }
+
     uniqueIIDs <- unique(c(highRelated$IID1, highRelated$IID2))
 
-    fail_other <- NULL
+    failIDs_other <- NULL
     if (!is.null(otherCriterion)) {
         otherCriterionThDirection <- match.arg(otherCriterionThDirection)
         if (!(otherCriterionMeasure %in% names(otherCriterion))) {
-            stop(paste("Column", otherCriterionMeasure, "not found!"))
+            stop(paste("Column", otherCriterionMeasure,
+                       "for otherCriterion not found!"))
         }
         if (!(otherCriterionIID %in% names(otherCriterion))) {
-            stop(paste("Column", otherCriterionIID, "not found!"))
+            stop(paste("Column", otherCriterionIID,
+                       "for otherCriterion not found!"))
         }
         names(otherCriterion)[names(otherCriterion) ==
                                   otherCriterionMeasure] <- "M"
         names(otherCriterion)[names(otherCriterion) ==
                                   otherCriterionIID] <- "IID"
 
+        if (any(!uniqueIIDs %in% otherCriterion$IID)) {
+            stop("Not all IIDs provided in relatedness are contained in" ,
+                 "otherCriterion")
+        }
         fail_other <- apply(highRelated, 1, function(pair) {
             failID1 <- NULL
             failID2 <- NULL
@@ -238,25 +279,38 @@ relatednessFilter <- function(relatedness, otherCriterion=NULL,
             if (two) failID2 <- pair[2]
             return(c(failID1, failID2))
         })
-        fail_other <- unique(unlist(fail_other))
+        failIDs_other <- unique(unlist(fail_other))
 
         # remove samples that fail other criterion
-        highRelated <- highRelated[!(highRelated$IID1 %in% fail_other |
-                                         highRelated$IID2 %in% fail_other), ]
+        highRelated <- highRelated[!(highRelated$IID1 %in% failIDs_other |
+                                         highRelated$IID2 %in% failIDs_other), ]
+        if (nrow(highRelated) == 0) {
+            if (verbose) {
+                message("Relatedness cannot be evaluated as all individuals ",
+                        "involved fail due to otherCriterion")
+            }
+            return(list(relatednessFails=NULL, failIDs=NULL))
+        }
         uniqueIIDs <- unique(c(highRelated$IID1, highRelated$IID2))
     }
     # all samples with related individuals
     allRelated <- c(highRelated$IID1, highRelated$IID2)
 
     # Further selection of samples with relatives in cohort
-    duplicateIDs <- unique(allRelated[duplicated(allRelated)])
+    multipleRelative <- unique(allRelated[duplicated(allRelated)])
+    singleRelative <- uniqueIIDs[!uniqueIIDs %in% multipleRelative]
 
-    if (length(duplicateIDs) == 0) {
-        # Only one related samples per individual
+    highRelatedMultiple <- highRelated[highRelated$IID1 %in% multipleRelative |
+                                        highRelated$IID2 %in% multipleRelative,]
+    highRelatedSingle <- highRelated[highRelated$IID1 %in% singleRelative &
+                                       highRelated$IID2 %in% singleRelative,]
+
+    # Only one related samples per individual
+    if(length(singleRelative) != 0) {
         if (!is.null(otherCriterion)) {
             # choose one with best otherCriterion, if equal arbitrarily return
             # the first one as fail
-            failIDs <- apply(highRelated, 1, function(pair) {
+            failIDs_single <- apply(highRelatedSingle, 1, function(pair) {
                 one_two <- evaluateDirection(
                     otherCriterion$M[otherCriterion$IID == pair[1]],
                     otherCriterion$M[otherCriterion$IID == pair[2]],
@@ -272,31 +326,58 @@ relatednessFilter <- function(relatedness, otherCriterion=NULL,
             })
         } else {
             # randomly choose one to exclude
-            failIDs <- highRelated[1,]
+            failIDs_single <- highRelatedSingle[,1]
         }
     } else {
-        # Get all related samples per individual
-        relatedPerID <- lapply(duplicateIDs, function(x) {
-            tmp <- highRelated[rowSums(cbind(highRelated$IID1 %in% x,
-                                             highRelated$IID2 %in% x)) != 0,1:2]
+        failIDs_single <- NULL
+    }
+
+    if(length(multipleRelative) != 0) {
+        relatedPerID <- lapply(multipleRelative, function(x) {
+            tmp <- highRelatedMultiple[rowSums(
+                cbind(highRelatedMultiple$IID1 %in% x,
+                      highRelatedMultiple$IID2 %in% x)) != 0,1:2]
             rel <- unique(unlist(tmp))
             return(rel)
         })
-        names(relatedPerID) <- duplicateIDs
+        names(relatedPerID) <- multipleRelative
 
-        # Find all non-related samples from family structures
-        nonRelated <- lapply(relatedPerID, function(rel) {
-            pairwise <- t(combn(rel, 2))
-            combination <- (pairwise[,1] %in% highRelated$IID1 &
-                            pairwise[,2] %in% highRelated$IID2) |
-                            (pairwise[,2] %in% highRelated$IID1 &
-                            pairwise[,1] %in% highRelated$IID2)
-            return(pairwise[!combination,])
+        keepIDs_multiple <- lapply(relatedPerID, function(x) {
+            pairwise <- t(combn(x, 2))
+            index <- (highRelatedMultiple$IID1 %in% pairwise[,1] &
+                          highRelatedMultiple$IID2 %in% pairwise[,2]) |
+                (highRelatedMultiple$IID1 %in% pairwise[,2] &
+                     highRelatedMultiple$IID2 %in% pairwise[,1])
+            combination <- highRelatedMultiple[index,]
+            combination_graph <- igraph::graph_from_data_frame(combination,
+                                                               directed=FALSE)
+            all_iv_set <- igraph::ivs(combination_graph)
+            length_iv_set <- sapply(all_iv_set, function(x) length(x))
+
+            if (all(length_iv_set == 1)) {
+                # check how often they occurr elsewhere
+                occurrence <- sapply(x, function(id) {
+                    sum(sapply(relatedPerID, function(idlist) id %in% idlist))
+                })
+                # if occurrence the same everywhere, pick the first, else keep
+                # the one with minimum occurrence elsewhere
+                if (length(unique(occurrence)) == 1) {
+                    nonRelated <- sort(x)[1]
+                } else {
+                    nonRelated <- names(occurrence)[which.min(occurrence)]
+                }
+            } else {
+                nonRelated <- all_iv_set[which.max(length_iv_set)]
+            }
+            return(nonRelated)
         })
-        nonRelated <- unique(unlist(nonRelated))
-        failIDs <- duplicateIDs[!duplicateIDs %in% nonRelated]
+        keepIDs_multiple <- unique(unlist(keepIDs_multiple))
+        failIDs_multiple <- c(multipleRelative[!multipleRelative %in%
+                                                   keepIDs_multiple])
+    } else {
+        failIDs_multiple <- NULL
     }
-    allFailIIDs <- c(failIDs, fail_other)
+    allFailIIDs <- c(failIDs_single, failIDs_multiple, failIDs_other)
     relatednessFails <- lapply(allFailIIDs, function(id) {
         fail_inorder <- relatedness_original$IID1 == id &
             relatedness_original$M > relatednessTh
@@ -304,16 +385,12 @@ relatednessFilter <- function(relatedness, otherCriterion=NULL,
             relatedness_original$M > relatednessTh
         if (any(fail_inreverse)) {
             inreverse <- relatedness_original[fail_inreverse, ]
-            if (relatednessFID2 %in% names(relatedness_original)) {
-                id1 <-
-                    c(which(colnames(relatedness_original) == relatednessIID1),
-                      which(colnames(relatedness_original) == relatednessFID1))
-                id2 <-
-                    c(which(colnames(relatedness_original) == relatednessIID2),
-                      which(colnames(relatedness_original) == relatednessFID2))
+            if (!is.null(relatednessFID2)) {
+                id1 <- c(iid1_index,fid1_index)
+                id2 <- c(iid2_index,fid2_index)
             } else {
-                id1 <- which(colnames(relatedness_original) == relatednessIID1)
-                id2 <- which(colnames(relatedness_original) == relatednessIID2)
+                id1 <- iid1_index
+                id2 <- iid2_index
             }
             inreverse[,c(id1, id2)] <- inreverse[,c(id2, id1)]
             names(inreverse) <- relatedness_names
@@ -331,12 +408,14 @@ relatednessFilter <- function(relatedness, otherCriterion=NULL,
     } else {
         names(relatednessFails) <- relatedness_names
         rownames(relatednessFails) <- 1:nrow(relatednessFails)
-        uniqueFails <- relatednessFails[!duplicated(relatednessFails$IID1),]
-        if (relatednessFID2 %in% names(relatedness_original)) {
-            failIDs <- data.frame(FID=uniqueFails$FID1,
-                              IID=uniqueFails$IID1)
+        uniqueFails <- relatednessFails[!duplicated(relatednessFails[,iid1_index]),]
+        if (!is.null(relatednessFID2)) {
+            failIDs <- data.frame(FID=uniqueFails[,fid1_index],
+                              IID=uniqueFails[,iid1_index],
+                              stringsAsFactors=FALSE)
         } else {
-            failIDs <- data.frame(IID=uniqueFails$IID1)
+            failIDs <- data.frame(IID=uniqueFails[,iid1_index],
+                                  stringsAsFactors=FALSE)
         }
     }
     return(list(relatednessFails=relatednessFails, failIDs=failIDs))
